/*
    ES Knot
    Copyright 2009 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include "knot2.h"
#include <GLES2/gl2.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifndef NO_EGL
#include <EGL/egl.h>
#endif

#define USE_FBO
#define USE_HARDWARE

#ifdef RENAME_OES_2_CLASS
#define Knot Knot2
#define PbufferRTT PbufferRTT2
#define RadialBlur RadialBlur2
#define MotionBlur MotionBlur2
#endif

void debug(const char *str);

#define NO_CELCOORDS
#define NO_OUTLINE_VERTICES
#define ADJUST_BUMP_NORMALS
#include "knotbase.cpp"

int compileShader(GLuint shader, const char* source) {
	glShaderSource(shader, 1, &source, NULL);
	glCompileShader(shader);
	GLint shaderCompiled;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &shaderCompiled);
	if (!shaderCompiled) {
		GLint logLength;
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &logLength);
		char* infoLog = (char*) malloc(logLength);
		GLint charsWritten;
		glGetShaderInfoLog(shader, logLength, &charsWritten, infoLog);
		debug(infoLog);
		free(infoLog);
		return 0;
	}
	return 1;
}

int linkShader(GLuint program, GLuint vertShader, GLuint fragShader) {
	glAttachShader(program, fragShader);
	glAttachShader(program, vertShader);
	glLinkProgram(program);
	GLint linked;
	glGetProgramiv(program, GL_LINK_STATUS, &linked);
	if (!linked) {
		GLint logLength;
		glGetProgramiv(program, GL_INFO_LOG_LENGTH, &logLength);
		char* infoLog = (char*) malloc(logLength);
		GLint charsWritten;
		glGetProgramInfoLog(program, logLength, &charsWritten, infoLog);
		debug(infoLog);
		free(infoLog);
		return 0;
	}
	return 1;
}

bool initProgram(GLuint& program, GLuint& vertShader, GLuint& fragShader, const char* vertSource, const char* fragSource) {
	program = vertShader = fragShader = 0;
	vertShader = glCreateShader(GL_VERTEX_SHADER);
	if (!compileShader(vertShader, vertSource))
		return false;
	fragShader = glCreateShader(GL_FRAGMENT_SHADER);
	if (!compileShader(fragShader, fragSource))
		return false;
	program = glCreateProgram();
	if (!linkShader(program, vertShader, fragShader))
		return false;
	return true;
}

void clearProgram(GLuint& program, GLuint& vertShader, GLuint& fragShader) {
	glDeleteProgram(program);
	glDeleteShader(vertShader);
	glDeleteShader(fragShader);
	program = vertShader = fragShader = 0;
}

void matrixMult(float* out, const float* in1, const float* in2) {
	int i, j, k;
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 4; j++) {
			float sum = 0;
			for (k = 0; k < 4; k++)
				sum += in1[4*k+j]*in2[4*i+k];
			out[4*i+j] = sum;
		}
	}
}

void matrixMultVect(float* out, const float* matrix, const float* vect) {
	int i, j;
	for (i = 0; i < 4; i++) {
		float sum = 0;
		for (j = 0; j < 4; j++)
			sum += matrix[4*j+i]*vect[j];
		out[i] = sum;
	}
}

void vectMult(float* out, const float* in1, const float* in2) {
	for (int i = 0; i < 4; i++)
		out[i] = in1[i] * in2[i];
}

void vectCopy(float* out, const float* in) {
	memcpy(out, in, sizeof(float)*4);
}

void matrixLoadIdentity(float* matrix) {
	memset(matrix, 0, sizeof(float)*16);
	for (int i = 0; i < 4; i++)
		matrix[4*i+i] = 1;
}

void matrixCopy(float* out, const float* in) {
	memcpy(out, in, sizeof(float)*16);
}

void matrixMultRight(float* out, const float* in) {
	float temp[16];
	matrixMult(temp, out, in);
	matrixCopy(out, temp);
}

void matrixRotation(float* matrix, float angle, float x, float y, float z) {
	float invlen = 1/sqrt(x*x + y*y + z*z);
	x *= invlen;
	y *= invlen;
	z *= invlen;
	float c = cos(M_PI*angle/180);
	float s = sin(M_PI*angle/180);
	float rotMatrix[] = { // transposed
		x*x*(1-c)+c,	x*y*(1-c)-z*s,	x*z*(1-c)+y*s,	0,
		x*y*(1-c)+z*s,	y*y*(1-c)+c,	y*z*(1-c)-x*s,	0,
		x*z*(1-c)-y*s,	y*z*(1-c)+x*s,	z*z*(1-c)+c,	0,
		0,		0,		0,		1
	};
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			matrix[4*i+j] = rotMatrix[4*j+i];
}

void matrixMultRotate(float* matrix, float angle, float x, float y, float z) {
	float rot[16];
	matrixRotation(rot, angle, x, y, z);
	matrixMultRight(matrix, rot);
}

void matrixCopyRotation(float* out, const float* in) {
	matrixLoadIdentity(out);
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			out[4*i+j] = in[4*i+j];
}

void matrixResetRotation(float* matrix) {
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			matrix[4*i+j] = i == j ? 1 : 0;
}


// Only works as long as the upper left part is a pure rotation
void matrixInv(float* out, const float* in) {
	// in' = (T*R)' = R'*T'
	float rotation[16], translation[16];
	matrixLoadIdentity(rotation);
	matrixLoadIdentity(translation);
	// the rotational part from in, transposed (which we assume is the same as the inverse)
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			rotation[4*i+j] = in[4*j+i];
	// inverse translation
	for (int i = 0; i < 3; i++)
		translation[4*3+i] = -in[4*3+i];
	matrixMult(out, rotation, translation);
}

void matrixFrustum(float* matrix, float left, float right, float bottom, float top, float zNear, float zFar) {
	float projMatrix[] = { // note that specifying a matrix in this form makes it transposed
		2*zNear / (right - left),	0,				 (right + left) / (right - left),	0,
		0,				2*zNear / (top - bottom),	 (top + bottom) / (top - bottom),	0,
		0,				0,				-(zFar + zNear) / (zFar - zNear),	-(2*zFar*zNear) / (zFar - zNear),
		0,				0,				-1,					0
	};
	// copy transposed
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			matrix[4*i+j] = projMatrix[4*j+i];
}

void matrixOrtho(float* matrix, float left, float right, float bottom, float top, float zNear, float zFar) {
	float projMatrix[] = { // note that specifying a matrix in this form makes it transposed
		2 / (right - left),	0,			 0,			-(right + left) / (right - left),
		0,			2 / (top - bottom),	 0,			-(top + bottom) / (top - bottom),
		0,			0,			-2 / (zFar - zNear),	-(zFar + zNear) / (zFar - zNear),
		0,			0,			 0,			1
	};
	// copy transposed
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			matrix[4*i+j] = projMatrix[4*j+i];
}



#include "pbuffer.cpp"

const char* blendVertShaderSource = "\
attribute highp vec2 vertex;\
attribute highp vec2 texCoord;\
attribute lowp vec4 color;\
uniform mediump mat4 pmv;\
\
varying mediump vec2 texCoordVarying;\
varying lowp vec4 colorVarying;\
void main(void) {\
	texCoordVarying = texCoord;\
	colorVarying = color;\
	gl_Position = pmv * vec4(vertex, 0.0, 1.0);\
}";

const char* blendFragShaderSource = "\
uniform sampler2D texture;\
varying mediump vec2 texCoordVarying;\
varying lowp vec4 colorVarying;\
void main (void) {\
	gl_FragColor = vec4(vec3(texture2D(texture, texCoordVarying)), colorVarying.a);\
}";

class RadialBlur {
public:
	RadialBlur(int vpWidth, int vpHeight) {
		pbuffer = new PbufferRTT();
//		pbuffer->init(vpWidth, vpHeight);
		pbuffer->init(256, 256);
		textureWidth = pbuffer->getWidth();
		textureHeight = pbuffer->getHeight();
		initProgram(blendProg, blendVert, blendFrag, blendVertShaderSource, blendFragShaderSource);
		matrixOrtho(projMatrix, 0, 100, 0, 100, -1, 1);
	}
	~RadialBlur() {
		clearProgram(blendProg, blendVert, blendFrag);
		delete pbuffer;
	}
	void setLinear(bool linear) {
		pbuffer->bind();
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, linear ? GL_LINEAR : GL_NEAREST);
	}
	void setGlow(bool g) {
		glow = g;
	}

	void init(int& vpWidth, int& vpHeight) {
		origWidth = vpWidth;
		origHeight = vpHeight;

		pbuffer->activate();
		vpWidth = textureWidth;
		vpHeight = textureHeight;
	}
	void finish(int& vpWidth, int& vpHeight) {

		pbuffer->finish();
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);
		pbuffer->bind();


#define RADIALLAYERS 5

		TYPE tex[2*4*RADIALLAYERS];
		TYPE verts[2*4*RADIALLAYERS];
		TYPE colors[4*4*RADIALLAYERS];
		GLushort indices[6*RADIALLAYERS];
		static const GLushort rectindices[] = { 0, 1, 2, 0, 2, 3 };

		TYPE texWidth = F(vpWidth)/textureWidth;
		TYPE texHeight = F(vpHeight)/textureHeight;
		for (int i = 0; i < RADIALLAYERS; i++) {
			verts[2*(4*i+0)+0] = F(0);
			verts[2*(4*i+0)+1] = F(0);
			verts[2*(4*i+1)+0] = F(100);
			verts[2*(4*i+1)+1] = F(0);
			verts[2*(4*i+2)+0] = F(100);
			verts[2*(4*i+2)+1] = F(100);
			verts[2*(4*i+3)+0] = F(0);
			verts[2*(4*i+3)+1] = F(100);
			for (int j = 0; j < 6; j++)
				indices[6*i+j] = rectindices[j] + 4*i;
			TYPE alpha = F(1)/(i+1);
			if (i == 0 && glow)
				alpha = F(1);
			else
				alpha = F(1)/RADIALLAYERS;
			for (int j = 0; j < 4; j++) {
				colors[4*(4*i+j)+0] = F(1);
				colors[4*(4*i+j)+1] = F(1);
				colors[4*(4*i+j)+2] = F(1);
				colors[4*(4*i+j)+3] = alpha;
			}
			TYPE margin = MULT(F(0.35), F(i))/RADIALLAYERS;
			TYPE xmargin = MULT(texWidth, margin)/2;
			TYPE ymargin = MULT(texHeight, margin)/2;
			tex[2*(4*i+0)+0] = xmargin;
			tex[2*(4*i+0)+1] = ymargin;
			tex[2*(4*i+1)+0] = texWidth - xmargin;
			tex[2*(4*i+1)+1] = ymargin;
			tex[2*(4*i+2)+0] = texWidth - xmargin;
			tex[2*(4*i+2)+1] = texHeight - ymargin;
			tex[2*(4*i+3)+0] = xmargin;
			tex[2*(4*i+3)+1] = texHeight - ymargin;
		}


		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glDisable(GL_CULL_FACE);

		GLuint prog = blendProg;
		glUseProgram(prog);
		glUniformMatrix4fv(glGetUniformLocation(prog, "pmv"), 1, GL_FALSE, projMatrix);
		glUniform1i(glGetUniformLocation(prog, "texture"), 0);
		GLuint vertexAttr = glGetAttribLocation(prog, "vertex");
		GLuint texCoordAttr = glGetAttribLocation(prog, "texCoord");
		GLuint colorAttr = glGetAttribLocation(prog, "color");
		glEnableVertexAttribArray(vertexAttr);
		glVertexAttribPointer(vertexAttr, 2, GL_FLOAT, GL_FALSE, 0, verts);
		glEnableVertexAttribArray(texCoordAttr);
		glVertexAttribPointer(texCoordAttr, 2, GL_FLOAT, GL_FALSE, 0, tex);
		glEnableVertexAttribArray(colorAttr);
		glVertexAttribPointer(colorAttr, 4, GL_FLOAT, GL_FALSE, 0, colors);

		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glDrawElements(GL_TRIANGLES, 6*RADIALLAYERS, GL_UNSIGNED_SHORT, indices);

		glDisableVertexAttribArray(vertexAttr);
		glDisableVertexAttribArray(texCoordAttr);
		glDisableVertexAttribArray(colorAttr);

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);

		vpWidth = origWidth;
		vpHeight = origHeight;
	}
private:
	PbufferRTT* pbuffer;
	float projMatrix[16];
	GLuint blendProg, blendVert, blendFrag;
	int textureWidth, textureHeight;
	int origWidth, origHeight;
	bool glow;
};

class MotionBlur {
public:
	MotionBlur(int vpWidth, int vpHeight) {
		pbuffer1 = new PbufferRTT();
		pbuffer2 = new PbufferRTT();
//		pbuffer1->init(vpWidth, vpHeight);
//		pbuffer2->init(vpWidth, vpHeight);
		pbuffer1->init(256, 256);
		pbuffer2->init(256, 256);
		initProgram(blendProg, blendVert, blendFrag, blendVertShaderSource, blendFragShaderSource);
		matrixOrtho(projMatrix, 0, 100, 0, 100, -1, 1);
	}
	~MotionBlur() {
		clearProgram(blendProg, blendVert, blendFrag);
		delete pbuffer1;
		delete pbuffer2;
	}

	void init(int& vpWidth, int& vpHeight) {
		origWidth = vpWidth;
		origHeight = vpHeight;
		pbuffer1->activate();
		textureWidth = vpWidth = pbuffer1->getWidth();
		textureHeight = vpHeight = pbuffer1->getHeight();
	}
	void finish(int& vpWidth, int& vpHeight, unsigned int diff) {
		TYPE alpha = F(1 - exp(-float(diff)/130));

		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_CULL_FACE);

		pbuffer2->bind();

		TYPE tex[2*4];
		TYPE verts[2*4];
		TYPE colors[4*4];
		tex[2*0+0] = F(0);
		tex[2*0+1] = F(0);
		tex[2*1+0] = F(vpWidth)/textureWidth;
		tex[2*1+1] = F(0);
		tex[2*2+0] = F(vpWidth)/textureWidth;
		tex[2*2+1] = F(vpHeight)/textureHeight;
		tex[2*3+0] = F(0);
		tex[2*3+1] = F(vpHeight)/textureHeight;
		verts[2*0+0] = F(0);
		verts[2*0+1] = F(0);
		verts[2*1+0] = F(100);
		verts[2*1+1] = F(0);
		verts[2*2+0] = F(100);
		verts[2*2+1] = F(100);
		verts[2*3+0] = F(0);
		verts[2*3+1] = F(100);
		colors[4*0+0] = colors[4*0+1] = colors[4*0+2] = F(1);
		colors[4*1+0] = colors[4*1+1] = colors[4*1+2] = F(1);
		colors[4*2+0] = colors[4*2+1] = colors[4*2+2] = F(1);
		colors[4*3+0] = colors[4*3+1] = colors[4*3+2] = F(1);
		colors[4*0+3] = colors[4*1+3] = colors[4*2+3] = colors[4*3+3] = F(1) - alpha;
		static const GLushort indices[] = { 0, 1, 2, 0, 2, 3 };

		GLuint prog = blendProg;
		glUseProgram(prog);
		glUniformMatrix4fv(glGetUniformLocation(prog, "pmv"), 1, GL_FALSE, projMatrix);
		glUniform1i(glGetUniformLocation(prog, "texture"), 0);
		GLuint vertexAttr = glGetAttribLocation(prog, "vertex");
		GLuint texCoordAttr = glGetAttribLocation(prog, "texCoord");
		GLuint colorAttr = glGetAttribLocation(prog, "color");
		glEnableVertexAttribArray(vertexAttr);
		glVertexAttribPointer(vertexAttr, 2, GL_FLOAT, GL_FALSE, 0, verts);
		glEnableVertexAttribArray(texCoordAttr);
		glVertexAttribPointer(texCoordAttr, 2, GL_FLOAT, GL_FALSE, 0, tex);
		glEnableVertexAttribArray(colorAttr);
		glVertexAttribPointer(colorAttr, 4, GL_FLOAT, GL_FALSE, 0, colors);

		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);

		pbuffer1->finish();
		pbuffer1->bind();
		glViewport(0, 0, origWidth, origHeight);
		glDisable(GL_BLEND);
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, indices);

		glDisableVertexAttribArray(vertexAttr);
		glDisableVertexAttribArray(texCoordAttr);
		glDisableVertexAttribArray(colorAttr);

		glDisable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		vpWidth = origWidth;
		vpHeight = origHeight;

		PbufferRTT* tmp = pbuffer1;
		pbuffer1 = pbuffer2;
		pbuffer2 = tmp;
	}
	void clear() {
		pbuffer1->clear();
		pbuffer2->clear();
	}
private:
	int textureWidth, textureHeight;
	int origWidth, origHeight;
	float projMatrix[16];
	GLuint blendProg, blendVert, blendFrag;
	PbufferRTT* pbuffer1, *pbuffer2;
};

Knot::Knot(int w, int h) {
	width = w;
	height = h;
	updateBumps = false;
	doMotionBlur = false;
	doRadialBlur = false;
	morphing = false;
	rotating = true;
	gravitySet = false;

	texture = 0;
	celTexture = 0;

	motionBlur = NULL;
	radialBlur = NULL;

	precalc();

	untexturedProg = untexturedFrag = untexturedVert = 0;
	texSpecProg = texSpecVert = texSpecFrag = 0;
	outlineProg = outlineVert = outlineFrag = 0;
	celShadeProg = celShadeVert = celShadeFrag = 0;
	reinit(UNTEXTURED);
}

void Knot::setUpdateVertices(bool update) {
	updateBumps = update;
	setTexMode(texMode);
	if (!updateBumps)
		calcDynVerticesNormals(vertices, normals, ringcoords);
}

void Knot::setMotionBlur(bool blur) {
	doMotionBlur = blur;
	if (blur)
		motionBlur->clear();
}

void Knot::setRadialBlur(bool blur) {
	doRadialBlur = blur;
}

void Knot::setRadialGlow(bool g) {
	radialBlur->setGlow(g);
}

void Knot::setSize(int w, int h) {
	width = w;
	height = h;
	initProjection();
}

void Knot::reinit() {
	reinit(texMode);
}

void Knot::enableRotate(bool r) {
	rotating = r;
}

void Knot::enableGravity(bool g) {
	gravityEnabled = g;
	gravitySet = false;
}

void Knot::rotate(int x, int y) {
	float rot[16];
	matrixCopyRotation(rot, modelViewMatrix);
	matrixResetRotation(modelViewMatrix);

	int minDim = width;
	if (height < minDim)
		minDim = height;
	matrixMultRotate(modelViewMatrix, F(180)*x/minDim, F(0), F(1), F(0));
	matrixMultRotate(modelViewMatrix, F(180)*y/minDim, F(1), F(0), F(0));
	matrixMultRight(modelViewMatrix, rot);
}

const char* untexturedVertShaderSource = "\
attribute highp vec3 vertex;\
attribute highp vec3 normal;\
uniform mediump mat4 pmv;\
uniform mediump vec3 eye;\
\
uniform highp vec4 lightPos;\
uniform lowp vec4 diffuse;\
uniform lowp vec4 ambient;\
\
varying lowp vec4 color;\
varying mediump vec3 halfVec;\
varying mediump vec3 normalVarying;\
void main(void) {\
	highp vec3 vertexToLight = normalize(vec3(lightPos) - vertex);\
	float d = max(0.0, dot(vertexToLight, normal));\
	halfVec = normalize(vertexToLight + eye);\
	normalVarying = normal;\
	color = ambient + diffuse * d;\
	gl_Position = pmv * vec4(vertex, 1.0);\
}";

const char* untexturedFragShaderSource = "\
uniform mediump float shininess;\
uniform lowp vec4 specular;\
varying lowp vec4 color;\
varying mediump vec3 halfVec;\
varying mediump vec3 normalVarying;\
void main (void) {\
	mediump float s = pow(max(0.0, dot(normalize(halfVec), normalize(normalVarying))), shininess);\
	gl_FragColor = color + specular * s;\
}";

const char* texSpecVertShaderSource = "\
attribute highp vec3 vertex;\
attribute highp vec3 normal;\
attribute highp vec2 texcoords;\
uniform mediump mat4 pmv;\
uniform mediump vec3 eye;\
\
uniform highp vec4 lightPos;\
uniform lowp vec4 diffuse;\
uniform lowp vec4 ambient;\
uniform mediump float texOffsetY;\
\
varying lowp vec4 color;\
varying mediump vec3 halfVec;\
varying mediump vec3 normalVarying;\
varying mediump vec2 texCoordVarying;\
void main(void) {\
	highp vec3 vertexToLight = normalize(vec3(lightPos) - vertex);\
	float d = max(0.0, dot(vertexToLight, normal));\
	halfVec = normalize(vertexToLight + eye);\
	normalVarying = normal;\
	color = ambient + diffuse * d;\
	texCoordVarying = vec2(texcoords.s, texcoords.t + texOffsetY);\
	gl_Position = pmv * vec4(vertex, 1.0);\
}";

const char* texSpecFragShaderSource = "\
uniform mediump float shininess;\
uniform lowp vec4 specular;\
uniform sampler2D texture;\
varying lowp vec4 color;\
varying mediump vec3 halfVec;\
varying mediump vec3 normalVarying;\
varying mediump vec2 texCoordVarying;\
void main (void) {\
	mediump float s = pow(max(0.0, dot(normalize(halfVec), normalize(normalVarying))), shininess);\
	gl_FragColor = color * texture2D(texture, texCoordVarying) + specular * s;\
}";

const char* outlineVertShaderSource = "\
attribute highp vec3 vertex;\
attribute highp vec3 normal;\
uniform mediump mat4 pmv;\
uniform mediump float add;\
\
void main(void) {\
	gl_Position = pmv * vec4(vertex + add*normal, 1.0);\
}";

const char* outlineFragShaderSource = "\
void main (void) {\
	gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);\
}";

const char* celShadeVertShaderSource = "\
attribute highp vec3 vertex;\
attribute highp vec3 normal;\
uniform mediump mat4 pmv;\
\
uniform highp vec4 lightPos;\
\
varying mediump vec2 texCoordVarying;\
void main(void) {\
	highp vec3 vertexToLight = normalize(vec3(lightPos) - vertex);\
	float d = dot(vertexToLight, normal);\
	texCoordVarying = vec2(1.0 - d, 0.0);\
	gl_Position = pmv * vec4(vertex, 1.0);\
}";

const char* celShadeFragShaderSource = "\
uniform sampler2D texture;\
varying mediump vec2 texCoordVarying;\
void main (void) {\
	gl_FragColor = texture2D(texture, texCoordVarying);\
}";


void Knot::clearContext() {
	glUseProgram(0);
	glBindTexture(GL_TEXTURE_2D, 0);
	clearProgram(untexturedProg, untexturedVert, untexturedFrag);
	clearProgram(texSpecProg, texSpecVert, texSpecFrag);
	clearProgram(outlineProg, outlineVert, outlineFrag);
	clearProgram(celShadeProg, celShadeVert, celShadeFrag);
	glDeleteTextures(1, &texture);
	glDeleteTextures(1, &celTexture);
	texture = 0;
	celTexture = 0;
	delete radialBlur;
	delete motionBlur;
	radialBlur = NULL;
	motionBlur = NULL;
}

void Knot::reinit(TexMode mode) {
	texMode = mode;
	initProgram(untexturedProg, untexturedVert, untexturedFrag, untexturedVertShaderSource, untexturedFragShaderSource);
	initProgram(texSpecProg, texSpecVert, texSpecFrag, texSpecVertShaderSource, texSpecFragShaderSource);
	initProgram(outlineProg, outlineVert, outlineFrag, outlineVertShaderSource, outlineFragShaderSource);
	initProgram(celShadeProg, celShadeVert, celShadeFrag, celShadeVertShaderSource, celShadeFragShaderSource);

	genTexture();
	genCelTexture();

	radialBlur = new RadialBlur(width, height);
	motionBlur = new MotionBlur(width, height);

	initProjection();

	matrixLoadIdentity(modelViewMatrix);
	modelViewMatrix[14] = -10;

	lasttime = (unsigned int)-1;
	texOffsetY = 0;
	bumpOffset = 0;
}

void Knot::initProjection() {
	float zn = 5;
	float planeSize = zn*tan((54/2)*M_PI/180);
	float fw, fh;
	if (width < height) {
		fw = planeSize;
		fh = fw*height/width;
	} else {
		fh = planeSize;
		fw = fh*width/height;
	}

	matrixFrustum(projMatrix, F(-fw), F(fw), F(-fh), F(fh), F(zn), F(15));
}

static const TYPE lightPos[] = { F(-3), F(10), F(0), F(1) };
static const TYPE lightAmbient[] = { F(0.2), F(0.2), F(0.2), F(1) };
static const TYPE lightDiffuse[] = { F(1), F(1), F(1), F(1) };

void Knot::setTexMode(TexMode mode) {
	texMode = mode;
}

Knot::~Knot() {
	clearContext();
}

bool Knot::getMatrixSupported() {
	return true;
}

void Knot::startMorph() {
	if (morphing) return;
	morphing = true;
	morphStart = lasttime;
}

void Knot::draw(unsigned int msec) {
	if (lasttime == (unsigned int)-1)
		lasttime = msec;
	unsigned int diff = msec - lasttime;
	lasttime = msec;

	TYPE *verts = vertices;
	TYPE *norms = normals;

	texOffsetY += F(diff)/30000;
	bumpOffset += F(diff)/4000;
#ifdef USING_FIXED
	texOffsetY &= 0x0000FFFF;
	bumpOffset &= 0x0000FFFF;
#else
	if (texOffsetY > 1) texOffsetY -= int(texOffsetY);
	if (bumpOffset > 1) bumpOffset -= int(bumpOffset);
#endif

	if (morphing) {
#define MORPHLENGTH 1200
		if (msec - morphStart > MORPHLENGTH) {
			morphing = false;
			if (ringcoords == ringcoords23)
				ringcoords = ringcoords25;
			else
				ringcoords = ringcoords23;
			calcDynVerticesNormals(vertices, normals, ringcoords);
		} else {
			TYPE *next = ringcoords23;
			if (ringcoords == ringcoords23)
				next = ringcoords25;
			morphMatrices(ringcoordstemp, ringcoords, next, F(msec-morphStart)/MORPHLENGTH);
			calcDynVerticesNormals(verts, normals, ringcoordstemp);
		}
	} else if (updateBumps)
		calcDynVerticesNormals(vertices, normals, ringcoords);

	if (rotating) {
		matrixMultRotate(modelViewMatrix, 90.0*diff/1000, sin(msec*M_PI/3000), sin(msec*M_PI/600), cos(msec*M_PI/900));
	} else if (gravityEnabled && gravitySet) {
		AVG3(curGravity, curGravity, targetGravity);
		NORMALIZE3(curGravity);
		rotateGravity();
	}

	int vpwidth = width;
	int vpheight = height;

	if (doMotionBlur)
		motionBlur->init(vpwidth, vpheight);

	if (doRadialBlur)
		radialBlur->init(vpwidth, vpheight);

	glViewport(0, 0, vpwidth, vpheight);

	float pmv[16];
	matrixMult(pmv, projMatrix, modelViewMatrix);
	float imv[16];
	matrixInv(imv, modelViewMatrix);
	float lightObjSpace[4];
	float eye[4] = { 0, 0, 1, 0 };
	float eyeObjSpace[4];
	matrixMultVect(lightObjSpace, imv, lightPos);
	matrixMultVect(eyeObjSpace, imv, eye);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	if (texMode == CELSHADED) {
		glClearColor(1, 1, 1, 1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		GLuint prog = outlineProg;
		glUseProgram(prog);
		glUniformMatrix4fv(glGetUniformLocation(prog, "pmv"), 1, GL_FALSE, pmv);
		glUniform1f(glGetUniformLocation(prog, "add"), OUTLINEADD);
		int vertexAttr = glGetAttribLocation(prog, "vertex");
		int normalAttr = glGetAttribLocation(prog, "normal");
		glEnableVertexAttribArray(vertexAttr);
		glVertexAttribPointer(vertexAttr, 3, GL_FLOAT, GL_FALSE, 0, verts);
		glEnableVertexAttribArray(normalAttr);
		glVertexAttribPointer(normalAttr, 3, GL_FLOAT, GL_FALSE, 0, norms);
		glCullFace(GL_FRONT);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);
		glDisableVertexAttribArray(vertexAttr);
		glDisableVertexAttribArray(normalAttr);

		prog = celShadeProg;
		glUseProgram(prog);
		glUniformMatrix4fv(glGetUniformLocation(prog, "pmv"), 1, GL_FALSE, pmv);
		glUniform4fv(glGetUniformLocation(prog, "lightPos"), 1, lightObjSpace);
		glBindTexture(GL_TEXTURE_2D, celTexture);
		glUniform1i(glGetUniformLocation(prog, "texture"), 0);
		vertexAttr = glGetAttribLocation(prog, "vertex");
		normalAttr = glGetAttribLocation(prog, "normal");
		glEnableVertexAttribArray(vertexAttr);
		glVertexAttribPointer(vertexAttr, 3, GL_FLOAT, GL_FALSE, 0, verts);
		glEnableVertexAttribArray(normalAttr);
		glVertexAttribPointer(normalAttr, 3, GL_FLOAT, GL_FALSE, 0, norms);
		glCullFace(GL_BACK);
		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);
		glDisableVertexAttribArray(vertexAttr);
		glDisableVertexAttribArray(normalAttr);
	} else {
		glClearColor(0, 0, 0, 1);
		float materialColor[4] = { 0.1, 0.1, 1, 1 };
		float materialAmbient[4] = { 0.2, 0.2, 0.2, 1 };
		float materialDiffuse[4] = { 0.8, 0.8, 0.8, 1 };
		float lightSpecular[4] = { 1, 1, 1, 1 };
		float materialSpecular[4] = { 0.8, 0.8, 0.8, 1 };
		float zero[4] = { 0, 0, 0, 1 };

		float ambient[4], diffuse[4], specular[4];
		vectMult(specular, materialSpecular, lightSpecular);
		GLuint prog = untexturedProg;
		if (texMode == TEXSPEC || texMode == PLAINTEX) {
			vectMult(ambient, materialAmbient, lightAmbient);
			vectMult(diffuse, materialDiffuse, lightDiffuse);
			prog = texSpecProg;
		} else {
			vectMult(ambient, materialColor, lightAmbient);
			vectMult(diffuse, materialColor, lightDiffuse);
		}

		glUseProgram(prog);
		glUniformMatrix4fv(glGetUniformLocation(prog, "pmv"), 1, GL_FALSE, pmv);
		glUniform4fv(glGetUniformLocation(prog, "lightPos"), 1, lightObjSpace);
		glUniform3fv(glGetUniformLocation(prog, "eye"), 1, eyeObjSpace);
		glUniform4fv(glGetUniformLocation(prog, "diffuse"), 1, diffuse);
		glUniform4fv(glGetUniformLocation(prog, "ambient"), 1, ambient);
		if (texMode == PLAINTEX)
			glUniform4fv(glGetUniformLocation(prog, "specular"), 1, zero);
		else
			glUniform4fv(glGetUniformLocation(prog, "specular"), 1, specular);
		glUniform1f(glGetUniformLocation(prog, "shininess"), 35);

		int vertexAttr = glGetAttribLocation(prog, "vertex");
		int normalAttr = glGetAttribLocation(prog, "normal");
		glEnableVertexAttribArray(vertexAttr);
		glVertexAttribPointer(vertexAttr, 3, GL_FLOAT, GL_FALSE, 0, verts);
		glEnableVertexAttribArray(normalAttr);
		glVertexAttribPointer(normalAttr, 3, GL_FLOAT, GL_FALSE, 0, norms);
		int texCoordAttr = -1;
		if (texMode == TEXSPEC || texMode == PLAINTEX) {
			glUniform1f(glGetUniformLocation(prog, "texOffsetY"), MULTY*texOffsetY);
			glBindTexture(GL_TEXTURE_2D, texture);
			glUniform1i(glGetUniformLocation(prog, "texture"), 0);
			texCoordAttr = glGetAttribLocation(prog, "texcoords");
			glEnableVertexAttribArray(texCoordAttr);
			glVertexAttribPointer(texCoordAttr, 2, GL_FLOAT, GL_FALSE, 0, texcoords);
		} else {
		}
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glDrawElements(GL_TRIANGLES, 6*RINGS*POINTS, GL_UNSIGNED_SHORT, indices);
		glDisableVertexAttribArray(vertexAttr);
		glDisableVertexAttribArray(normalAttr);
		if (texCoordAttr >= 0)
			glDisableVertexAttribArray(texCoordAttr);
	}

	if (doRadialBlur)
		radialBlur->finish(vpwidth, vpheight);

	if (doMotionBlur)
		motionBlur->finish(vpwidth, vpheight, diff);
}

void Knot::setGravity(TYPE x, TYPE y, TYPE z) {
	targetGravity[0] = x;
	targetGravity[1] = y;
	targetGravity[2] = z;

	if (!gravitySet) {
		COPY3(curGravity, targetGravity);
		complementVect[0] = F(1);
		complementVect[1] = F(0);
		complementVect[2] = F(0);
		extraRotInited = false;
	}
	gravitySet = true;
}

void Knot::rotateGravity() {
	TYPE thirdVect[3];

	cross3(thirdVect, curGravity, complementVect);
	NORMALIZE3(thirdVect);
	cross3(complementVect, thirdVect, curGravity);
	NORMALIZE3(complementVect);

	TYPE matrix[] = {
		complementVect[0], complementVect[1], complementVect[2], F(0),
		thirdVect[0], thirdVect[1], thirdVect[2], F(0),
		curGravity[0], curGravity[1], curGravity[2], F(0),
		modelViewMatrix[12], modelViewMatrix[13], modelViewMatrix[14], F(1)
	};

	if (!extraRotInited) {
		TYPE curRot[16];
		matrixCopyRotation(curRot, modelViewMatrix);

		TYPE startPos[16];
		TYPE invStartPos[16];
		matrixCopyRotation(startPos, matrix);
		matrixInv(invStartPos, startPos);

		matrixMult(extraRotMatrix, invStartPos, curRot);
		extraRotInited = true;
	}
	matrixMult(modelViewMatrix, matrix, extraRotMatrix);
}

KnotInterface* createKnot2(int w, int h) {
	return new Knot(w, h);
}


