/*
    ES Knot
    Copyright 2009 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define printGLErrors() do { int err = glGetError(); if (err) { char buf[50]; sprintf(buf, "%d: %x\n", __LINE__, err); debug(buf); } } while (0)
#define printEGLErrors() do { int err = eglGetError(); if (err != EGL_SUCCESS) { char buf[50]; sprintf(buf, "egl: %d: %x\n", __LINE__, err); debug(buf); } } while (0)

#define T_RAD1 2.28
#define T_RAD2 1.00
#define RADIUS F(0.7)
#define BUMPSIZE 0.20
#define OUTLINEADD 0.11
#define BUMPS 7

#define SUB3(a, b)		\
	do {			\
		a[0] -= b[0];	\
		a[1] -= b[1];	\
		a[2] -= b[2];	\
	} while (0);
#ifdef USING_FIXED
#define NORMALIZE3(v)								\
	do {					\
		float invlen = 1/sqrt(float(MULT(v[0], v[0]) + MULT(v[1], v[1]) + MULT(v[2], v[2]))/(1<<16));	\
		v[0] = TYPE(v[0] * invlen);	\
		v[1] = TYPE(v[1] * invlen);	\
		v[2] = TYPE(v[2] * invlen);	\
	} while (0);
#define AVG3(o, a, b)				\
	do {					\
		o[0] = (a[0] + b[0]) >> 1;	\
		o[1] = (a[1] + b[1]) >> 1;	\
		o[2] = (a[2] + b[2]) >> 1;	\
	} while (0);
#else
#define NORMALIZE3(v)								\
	do {									\
		TYPE invlen = 1/sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);	\
		v[0] *= invlen;							\
		v[1] *= invlen;							\
		v[2] *= invlen;							\
	} while (0);
#define AVG3(o, a, b)				\
	do {					\
		o[0] = (a[0] + b[0]) * 0.5;	\
		o[1] = (a[1] + b[1]) * 0.5;	\
		o[2] = (a[2] + b[2]) * 0.5;	\
	} while (0);
#endif
#define COPY3(a, b)		\
	do {			\
		a[0] = b[0];	\
		a[1] = b[1];	\
		a[2] = b[2];	\
	} while (0);

#define MIN(a, b) ((a) < (b) ? (a) : (b))

static int largerPOT(int x) {
	int out = 1;
	while (out < x)
		out <<= 1;
	return out;
}

static int smallerPOT(int x) {
	int out = largerPOT(x);
	if (out > x)
		out >>= 1;
	return out;
}

static GLuint genEmptyTexture(int &w, int &h) {
	GLuint tex;
	w = smallerPOT(w);
	h = smallerPOT(h);

	int max;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);
	if (w > max) w = max;
	if (h > max) h = max;

	while (true) {
		glGenTextures(1, &tex);
		glBindTexture(GL_TEXTURE_2D, tex);
		GLubyte *data = new GLubyte[w*h*3];
		memset(data, 0, sizeof(GLubyte)*w*h*3);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
		delete [] data;
		int err = glGetError();
		if (err != GL_OUT_OF_MEMORY)
			break;

		glDeleteTextures(1, &tex);
		if (w >= h)
			w /= 2;
		else
			h /= 2;
	}
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	return tex;
}

static void cross3(TYPE* out, const TYPE* v1, const TYPE* v2) {
	out[0] = MULT(v1[1], v2[2]) - MULT(v1[2], v2[1]);
	out[1] = MULT(v1[2], v2[0]) - MULT(v1[0], v2[2]);
	out[2] = MULT(v1[0], v2[1]) - MULT(v1[1], v2[0]);
}

void Knot::polar(float pos, TYPE *vect, int P, int Q) {
#define ZSCALE 1.35
	float r = T_RAD1 + T_RAD2*cos(2*M_PI*pos*Q);
	float theta = 2*M_PI*pos*P;
	float phi = 0.2*M_PI*sin(2*M_PI*pos*Q);
	vect[0] = F(r*cos(phi)*cos(theta));
	vect[1] = F(r*cos(phi)*sin(theta));
	vect[2] = F(ZSCALE*r*sin(phi));
}

static bool hasExtension(const char *str) {
	const char *extensions = (const char*) glGetString(GL_EXTENSIONS);
	int len = strlen(str);
	const char *ptr;
	while ((ptr = strstr(extensions, str)) != NULL) {
		if ((ptr == extensions || *(ptr-1) == ' ') && (*(ptr+len) == ' ' || *(ptr+len) == '\0'))
			return true;
		extensions = ptr+len;
	}
	return false;
}

static bool hasVersion(int major, int minor) {
	const char *version = (const char*) glGetString(GL_VERSION);
	while ((*version < '0' || *version > '9') && *version != '\0')
		version++;
	int v1, v2;
	if (sscanf(version, "%d.%d", &v1, &v2) != 2)
		return false;
	if (v1 < major) return false;
	if (v1 > major) return true;
	if (v2 < minor) return false;
	return true;
}

void Knot::genTexture() {
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	GLubyte* palette = new GLubyte[3*256];
	for (int i = 0; i < 256; i++) {
		GLubyte *b = &palette[3*i];
		b[0] = i;
		b[1] = i;
		b[2] = 255-i;
	}
	int size = 16;
	int level = 0;
	while (size >= 1) {
		GLubyte* pixels = new GLubyte[3*size*size];
		for (int y = 0; y < size; y++) {
			for (int x = 0; x < size; x++) {
				float fval = 0.5+0.5*sin(float(y)/size*2*M_PI + M_PI*sin(float(x)/size*2*M_PI));
				GLubyte val = GLubyte(255*fval);
				pixels[y*3*size + 3*x + 0] = palette[3*val + 0];
				pixels[y*3*size + 3*x + 1] = palette[3*val + 1];
				pixels[y*3*size + 3*x + 2] = palette[3*val + 2];
			}
		}
		glTexImage2D(GL_TEXTURE_2D, level, GL_RGB, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
		delete [] pixels;
		size /= 2;
		level++;
	}
#ifdef USE_HARDWARE
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
#else
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
#endif
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	delete [] palette;
}

void Knot::genCelTexture() {
	glGenTextures(1, &celTexture);
	glBindTexture(GL_TEXTURE_2D, celTexture);
	int steps = 2;
	static const float data[] = {
		1, 1, 1,
		0.60, 0.45, 0.45, 0.45,
		0.90, 0.20, 0.20, 0.20,
	};

	const float *dataptr = data;
	GLubyte r, g, b;
	r = int(255*(*dataptr++));
	g = int(255*(*dataptr++));
	b = int(255*(*dataptr++));

	float nextPos = 2;
	if (steps > 0)
		nextPos = *dataptr++;
	int next = 0;

	int size = 32;
	GLubyte* texdata = new GLubyte[3*size];
	for (int i = 0; i < size; i++) {
		float pos = float(i)/size;
		if (pos >= nextPos) {
			r = int(255*(*dataptr++));
			g = int(255*(*dataptr++));
			b = int(255*(*dataptr++));
			next++;
			if (next < steps)
				nextPos = *dataptr++;
			else
				nextPos = 2;

		}
		GLubyte *byte = &texdata[3*i];
		byte[0] = r;
		byte[1] = g;
		byte[2] = b;
	}
	int level = 0;
	while (size >= 1) {
		glTexImage2D(GL_TEXTURE_2D, level, GL_RGB, size, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, texdata);

		size /= 2;
		level++;
		if (size < 1) break;

		GLubyte *newtex = new GLubyte[3*size];
		for (int i = 0; i < size; i++) {
			GLubyte *out = &newtex[3*i];
			GLubyte *in = &texdata[2*3*i];
			out[0] = (int(in[0])+in[3])/2;
			out[1] = (int(in[1])+in[4])/2;
			out[2] = (int(in[2])+in[5])/2;
		}
		delete [] texdata;
		texdata = newtex;
	}
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	delete [] texdata;
}

void Knot::makeRingCoords(TYPE *coords, int P, int Q) {
	for (int i = 0; i < RINGS; i++) {
		TYPE *ring = &coords[16*i];
		TYPE *center = &ring[12];
		float pos = float(i)/RINGS;
		polar(pos, center, P, Q);
		TYPE *dir = &ring[8];
		polar(pos+float(1)/(5*RINGS), dir, P, Q);
		SUB3(dir, center);
		NORMALIZE3(dir);
		TYPE *xvect = &ring[0];
		TYPE *yvect = &ring[4];
		COPY3(xvect, center);
		xvect[2] = F(0);
		cross3(yvect, dir, xvect);
		cross3(xvect, yvect, dir);
		NORMALIZE3(xvect);
		NORMALIZE3(yvect);
		center[3] = F(1);
		dir[3] = F(0);
		xvect[3] = F(0);
		yvect[3] = F(0);
	}
}

void Knot::precalc() {
	for (int i = 0; i < SINETABLESIZE; i++)
		sinTable[i] = F(sin(2*M_PI*float(i)/SINETABLESIZE));

	makeRingCoords(ringcoords23, 2, 3);
	makeRingCoords(ringcoords25, 2, 5);
	ringcoords = ringcoords23;

	for (int i = 0; i < POINTS; i++) {
		TYPE* vertex = &ringvertices[4*i];
		float angle = 2*M_PI*i/POINTS;
		vertex[0] = F(cos(angle));
		vertex[1] = F(sin(angle));
		vertex[2] = F(1);
		vertex[3] = F(1);
	}

	for (int i = 0; i < RINGS; i++) {
		int cr = (POINTS+1)*i;
		int nr = (POINTS+1)*(i+1);
		for (int j = 0; j < POINTS; j++) {
			GLushort *offset = &indices[6*(POINTS*i+j)];
			int cp = j;
			int np = j+1;
			offset[0] = cr+cp;
			offset[1] = cr+np;
			offset[2] = nr+np;
			offset[3] = cr+cp;
			offset[4] = nr+np;
			offset[5] = nr+cp;
		}
	}

	bool bumps = updateBumps;
	updateBumps = false;
	calcDynVerticesNormals(vertices, normals, ringcoords);
	updateBumps = bumps;

#define MULTX 6
#define MULTY 100
	for (int i = 0; i <= RINGS; i++) {
		for (int j = 0; j <= POINTS; j++) {
			TYPE* texcoord = &texcoords[2*((POINTS+1)*i+j)];
			texcoord[0] = F(MULTX*float(j)/POINTS);
			texcoord[1] = F(MULTY*float(i)/RINGS);
		}
	}
#ifndef NO_CELCOORDS
	for (int i = 0; i <= RINGS; i++) {
		for (int j = 0; j <= POINTS; j++) {
			TYPE* coord = &celcoords[2*((POINTS+1)*i+j)];
			coord[0] = F(i)/RINGS;
			coord[1] = F(0);
		}
	}
#endif
}

#ifndef NO_CELCOORDS
void Knot::updateCelCoords(TYPE *verts, TYPE *norms) {
	TYPE light[4];
	getLightInverse(light);
	NORMALIZE3(light);
	TYPE *dir = light;

	for (int i = 0; i < RINGS; i++) {
//		TYPE* vertex = &verts[3*(POINTS+1)*i];
		TYPE* normal = &norms[3*(POINTS+1)*i];
		TYPE* coord = &celcoords[2*((POINTS+1)*i)];
		for (int j = 0; j < POINTS; j++) {
//			TYPE dir[4];
//			COPY3(dir, light);
//			SUB3(dir, vertex);
//			NORMALIZE3(dir);
			TYPE val = MULT(dir[0], normal[0]) + MULT(dir[1], normal[1]) + MULT(dir[2], normal[2]);
			if (val < F(0)) val = F(0);
			if (val > F(1)) val = F(1);
			coord[0] = F(1) - val;
			coord[1] = F(0);
//			vertex += 3;
			normal += 3;
			coord += 2;
		}
		memcpy(&celcoords[2*((POINTS+1)*i+POINTS)], &celcoords[2*((POINTS+1)*i+0)], sizeof(TYPE)*2);
	}
	memcpy(&celcoords[2*((POINTS+1)*RINGS)], &celcoords[0], sizeof(TYPE)*2*(POINTS+1));
}
#endif

void Knot::calcDynVertices() {
	for (int i = 0; i < RINGS; i++) {
		const TYPE* ring = &ringcoords[16*i];
		const TYPE* center = &ring[12];
//		const TYPE* dir = &ring[8];
		const TYPE* xvect = &ring[0];
		const TYPE* yvect = &ring[4];
		TYPE pos = F(i)/RINGS - bumpOffset;
		TYPE radius = RADIUS + MULT(F(BUMPSIZE), fastSin(BUMPS*pos));

		for (int j = 0; j < POINTS; j++) {
			TYPE* vertex = &vertices[3*((POINTS+1)*i+j)];
			const TYPE* origvertex = &ringvertices[4*j];
			const TYPE &x = origvertex[0];
			const TYPE &y = origvertex[1];
			vertex[0] = center[0] + MULT(xvect[0], MULT(radius, x)) + MULT(yvect[0], MULT(radius, y));
			vertex[1] = center[1] + MULT(xvect[1], MULT(radius, x)) + MULT(yvect[1], MULT(radius, y));
			vertex[2] = center[2] + MULT(xvect[2], MULT(radius, x)) + MULT(yvect[2], MULT(radius, y));
//			vertex[3] = F(1);
		}
		memcpy(&vertices[3*((POINTS+1)*i+POINTS)], &vertices[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
	}
	memcpy(&vertices[3*((POINTS+1)*RINGS)], &vertices[0], sizeof(TYPE)*3*(POINTS+1));
}

void Knot::calcDynVerticesNormals(TYPE *verts, TYPE *norms, const TYPE *coords) {
	for (int i = 0; i < RINGS; i++) {
		const TYPE* ring = &coords[16*i];
		const TYPE* center = &ring[12];
#ifdef ADJUST_BUMP_NORMALS
		const TYPE* dir = &ring[8];
#endif
		const TYPE* xvect = &ring[0];
		const TYPE* yvect = &ring[4];
		TYPE pos = F(i)/RINGS - bumpOffset;
		TYPE radius = RADIUS + MULT(F(BUMPSIZE), fastSin(BUMPS*pos));
#ifdef ADJUST_BUMP_NORMALS
		// 0.1 is just a magic constant, to compensate for the fact that
		// pos just goes from 0 to 1 while the real obj coordinates
		// change much more in the same range. Don't calculate any estimate
		// of the actual path length...
		TYPE derivative = MULT(F(BUMPSIZE*BUMPS*0.2), fastSin(BUMPS*pos + F(0.25)));
		if (!updateBumps) {
			radius = RADIUS;
			derivative = F(0);
		}
		float invNormalLength = 1/sqrt(1 + TOFLOAT(MULT(derivative, derivative)));
		TYPE normDirLength = TYPE(derivative*invNormalLength);
		TYPE normOutLength = F(invNormalLength);
#else
		if (!updateBumps)
			radius = RADIUS;
#endif

		for (int j = 0; j < POINTS; j++) {
			TYPE* vertex = &verts[3*((POINTS+1)*i+j)];
			const TYPE* origvertex = &ringvertices[4*j];
			const TYPE &x = origvertex[0];
			const TYPE &y = origvertex[1];
			vertex[0] = center[0] + MULT(xvect[0], MULT(radius, x)) + MULT(yvect[0], MULT(radius, y));
			vertex[1] = center[1] + MULT(xvect[1], MULT(radius, x)) + MULT(yvect[1], MULT(radius, y));
			vertex[2] = center[2] + MULT(xvect[2], MULT(radius, x)) + MULT(yvect[2], MULT(radius, y));
//			vertex[3] = F(1);

			TYPE* normal = &norms[3*((POINTS+1)*i+j)];
#ifdef ADJUST_BUMP_NORMALS
			const TYPE normx = MULT(normOutLength, x);
			const TYPE normy = MULT(normOutLength, y);
			const TYPE normz = -normDirLength;
			normal[0] = MULT(xvect[0], normx) + MULT(yvect[0], normy) + MULT(dir[0], normz);
			normal[1] = MULT(xvect[1], normx) + MULT(yvect[1], normy) + MULT(dir[1], normz);
			normal[2] = MULT(xvect[2], normx) + MULT(yvect[2], normy) + MULT(dir[2], normz);
#else
			normal[0] = MULT(xvect[0], x) + MULT(yvect[0], y);
			normal[1] = MULT(xvect[1], x) + MULT(yvect[1], y);
			normal[2] = MULT(xvect[2], x) + MULT(yvect[2], y);
#endif
		}
		memcpy(&verts[3*((POINTS+1)*i+POINTS)], &verts[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
		memcpy(&norms[3*((POINTS+1)*i+POINTS)], &norms[3*((POINTS+1)*i+0)], sizeof(TYPE)*3);
	}
	memcpy(&verts[3*((POINTS+1)*RINGS)], &verts[0], sizeof(TYPE)*3*(POINTS+1));
	memcpy(&norms[3*((POINTS+1)*RINGS)], &norms[0], sizeof(TYPE)*3*(POINTS+1));
}

TYPE Knot::fastSin(TYPE x) {
#ifdef USING_FIXED
	return sinTable[(x >> (16-SINETABLEMASKSIZE)) & SINETABLEMASK];
#else
	return sinTable[int(SINETABLESIZE*x) & SINETABLEMASK];
#endif
}

void Knot::morphMatrices(TYPE *out, const TYPE *first, const TYPE *second, TYPE pos) {
	TYPE pos1 = F(1) - pos;
	for (int i = 0; i < RINGS; i++) {
		const TYPE* matrix1 = &first[16*i];
		const TYPE* center1 = &matrix1[12];
//		const TYPE* dir1 = &matrix1[8];
		const TYPE* xvect1 = &matrix1[0];
		const TYPE* yvect1 = &matrix1[4];
		const TYPE* matrix2 = &second[16*i];
		const TYPE* center2 = &matrix2[12];
//		const TYPE* dir2 = &matrix2[8];
		const TYPE* xvect2 = &matrix2[0];
		const TYPE* yvect2 = &matrix2[4];
		TYPE* matrix3 = &out[16*i];
		TYPE* center3 = &matrix3[12];
		TYPE* dir3 = &matrix3[8];
		TYPE* xvect3 = &matrix3[0];
		TYPE* yvect3 = &matrix3[4];
#define MULTVECT(vect3, vect1, fact1, vect2, fact2)					\
		do {									\
			vect3[0] = MULT(vect1[0], fact1) + MULT(vect2[0], fact2);	\
			vect3[1] = MULT(vect1[1], fact1) + MULT(vect2[1], fact2);	\
			vect3[2] = MULT(vect1[2], fact1) + MULT(vect2[2], fact2);	\
			vect3[3] = vect1[3];						\
		} while (0)
		MULTVECT(center3, center1, pos1, center2, pos);
		MULTVECT(xvect3, xvect1, pos1, xvect2, pos);
		MULTVECT(yvect3, yvect1, pos1, yvect2, pos);
		cross3(dir3, xvect3, yvect3);
		cross3(yvect3, dir3, xvect3);
		NORMALIZE3(xvect3);
		NORMALIZE3(yvect3);
		NORMALIZE3(dir3);
		dir3[3] = 0;
	}
}

#ifndef NO_OUTLINE_VERTICES
void Knot::updateOutlineVertices(TYPE *outline, const TYPE *verts, const TYPE *norms) {
	int len = (RINGS + 1)*(POINTS + 1);
	for (int i = 0; i < len; i++) {
		const TYPE* vertex = &verts[3*i];
		const TYPE* normal = &norms[3*i];
		TYPE* out = &outline[3*i];
		out[0] = vertex[0] + MULT(F(OUTLINEADD), normal[0]);
		out[1] = vertex[1] + MULT(F(OUTLINEADD), normal[1]);
		out[2] = vertex[2] + MULT(F(OUTLINEADD), normal[2]);
	}
}
#endif


